/* Main function for root server.
   Copyright (C) 2003 Free Software Foundation, Inc.
   This file is part of the GNU Hurd.

   The GNU Hurd is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU Hurd is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307 USA.  */

#include <unistd.h>
#include <stdio.h>
#include <l4.h>

void
print_thread (l4_thread_id_t thread)
{
  if (l4_is_nil_thread (thread))
    printf ("nilthread");
  else if (l4_is_local_id (thread))
    printf ("Local 0x%x", thread.raw);
  else
    printf ("No 0x%x Version 0x%x", l4_thread_no (thread),
	    l4_version (thread));
}


void
print_thread_info (l4_thread_id_t thread)
{
  l4_thread_id_t dest = thread;
  l4_word_t control = 0;
  l4_word_t sp = 0;
  l4_word_t ip = 0;
  l4_word_t flags = 0;
  l4_word_t user_handle = 0;
  l4_thread_id_t pager = l4_nilthread;

  printf ("Thread ID: ");
  print_thread (l4_global_id_of (thread));
  printf (" ");
  print_thread (l4_local_id_of (thread));
  printf ("\n");

  printf ("Pager ID: ");
  print_thread (l4_pager_of (thread));
  printf ("\n");

  printf ("User Defined Handle: 0x%x\n", l4_user_defined_handle_of (thread));

  l4_exchange_registers (&dest, &control, &sp, &ip, &flags,
			 &user_handle, &pager);
  printf ("SP: 0x%x  IP: 0x%x  FLAGS: 0x%x\n", sp, ip, flags);
  printf ("State: %s, %s, %s\n",
	  (control & L4_XCHG_REGS_HALTED) ? "halted" : "running",
	  (control & L4_XCHG_REGS_SENDING) ? "sending" : "not sending",
	  (control & L4_XCHG_REGS_RECEIVING)
	  ? "receiving" : "not receiving");	   
  if (l4_same_threads (l4_myself (), l4_global_id_of (thread)))
    {
      printf ("Proc: %i  UserHandle: 0x%x  ExcHandler: ",
	      l4_processor_no (), l4_user_defined_handle ());
      print_thread (l4_exception_handler ());
      printf ("\n");
    }
}


#define STACK_SIZE 4096
char exception_handler_stack[STACK_SIZE];

void
exception_handler (void)
{
  int count = 0;

  while (count < 10)
    {
      l4_msg_tag_t tag;
      l4_thread_id_t from;
      l4_word_t mr[12];

      tag = l4_wait (&from);
      printf ("EXCEPTION HANDLER: Received message from: ");
      print_thread (from);
      printf ("\n");
      printf ("Tag: 0x%x", tag.raw);
      printf ("Label: %u  Untyped: %u  Typed: %u\n",
	      l4_label (tag), l4_untyped_words (tag), l4_typed_words (tag));

      l4_store_mrs (1, 12, mr);
      printf ("Succeeded: %u  Propagated: %u  Redirected: %u  Xcpu: %u\n",
	      l4_ipc_succeeded (tag), l4_ipc_propagated (tag),
	      l4_ipc_redirected (tag), l4_ipc_xcpu (tag));
      printf ("Error Code: 0x%x\n", l4_error_code ());
      printf ("EIP: 0x%x  EFLAGS: 0x%x  Exception: %u  ErrCode: %u\n",
	      mr[0], mr[1], mr[2], mr[3]);
      printf ("EDI: 0x%x  ESI: 0x%x  EBP: 0x%x  ESP: 0x%x\n",
	      mr[4], mr[5], mr[6], mr[7]);
      printf ("EAX: 0x%x  EBX: 0x%x  ECX: 0x%x  EDX: 0x%x\n",
	      mr[11], mr[8], mr[10], mr[9]);
    }

  printf ("EXCEPTION HANDLER: Too many exceptions.\n");

  while (1)
    l4_yield ();
}

extern void cls (void);

int
main (void)
{
  l4_thread_id_t thread;
  l4_word_t result;
  void *utcb;

  cls ();
  print_thread_info (l4_myself ());

  /* FIXME: This is not specified by the standard.  We don't know
     where and how much space we have for other thread's UTCB
     areas in the root server.  */
  utcb = (void *) ((l4_my_local_id ().raw & ~(l4_utcb_size () - 1))
    + l4_utcb_size ());
  thread = l4_global_id (l4_thread_no (l4_myself ()) + 1, 1);

  printf ("Creating exception handler thread at utcb 0x%x: ",
	  (l4_word_t) utcb);
  result = l4_thread_control (thread, l4_myself (), l4_myself (),
			      l4_nilthread, utcb);
  printf ("%s\n", result ? "successful" : "failed");
  /* Set the priority of the other thread to our priority.  Otherwise
     it is 100 and will never be scheduled as long as we are not in a
     blocking receive.  */
  l4_set_priority (thread, 255);
  l4_start_sp_ip (thread, ((l4_word_t) exception_handler_stack)
		  + sizeof (exception_handler_stack),
		  (l4_word_t) exception_handler);
  l4_set_exception_handler (thread);
  print_thread_info (l4_myself ());
  printf ("Sleeping for a second.\n");
  sleep (1);
  printf ("Crashing now.\n");

  asm ("hlt");

  while (1)
    l4_yield ();

  return 0;
}
